"""
Maybe more repsonse type will be done in the future
"""
import logging
import functools

from starlette.responses import (
    Response,
    HTMLResponse,
    PlainTextResponse,
    JSONResponse,
    RedirectResponse,
    StreamingResponse,
    FileResponse
)
from starlette.templating import Jinja2Templates

logger = logging.getLogger(__name__)

__all__ = [
    "register_type",
    "Response",
    "HTMLResponse",
    "PlainTextResponse",
    "JSONResponse",
    "RedirectResponse",
    "StreamingResponse",
    "FileResponse",
    "TemplateResponse",
]


TemplateResponse = Jinja2Templates(directory='templates').TemplateResponse


class AutoResponseType:

    type_map = {}

    @classmethod
    def register_type(cls, type_):
        def register_func(func):
            cls.type_map[type_] = func

            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)
            return wrapper
        return register_func

    @classmethod
    def automatic(cls, *args):

        # Response or Response subclass
        if isinstance(args[0], Response):
            return args[0]

        try:
            return cls.type_map[type(args[0])](*args)
        except KeyError:
            raise TypeError(f"Wrong response type: {type(args[0])}")


register_type = functools.partial(AutoResponseType.register_type)
automatic = functools.partial(AutoResponseType.automatic)


@register_type(dict)
def json_type(*args):
    if len(args) > 3:
        raise ValueError("The response cannot exceed three parameters.")

    # judge status code and headers
    try:
        if not isinstance(args[1], int):
            raise TypeError("The response status code must be integer.")

        if not isinstance(args[2], dict):
            raise TypeError("The response headers must be dictionary.")

    except IndexError:
        pass

    return JSONResponse(*args)


@register_type(str)
def text_type(*args):
    if len(args) > 3:
        raise ValueError("The response cannot exceed three parameters.")

    # judge status code and headers
    try:
        if not isinstance(args[1], int):
            raise TypeError("The response status code must be integer.")

        if not isinstance(args[2], dict):
            raise TypeError("The response headers must be dictionary.")

    except IndexError:
        pass

    return PlainTextResponse(*args)



import os
import json
import logging


__all__ = [
    "Config"
]

LOG_LEVELS = {
    "critical": logging.CRITICAL,
    "error": logging.ERROR,
    "warning": logging.WARNING,
    "info": logging.INFO,
    "debug": logging.DEBUG,
}


logger: logging.Logger = logging.getLogger("index")


class ConfigError(Exception):
    pass


class UpperDict:

    def __init__(self, data: dict):
        self.__dict = dict()
        for key in data.keys():
            self[key] = data[key]

    def __str__(self):
        indent = 4
        result = ["{"]

        def append(line):
            result.append(" "*indent + line)

        for key, value in self.__dict.items():
            if isinstance(value, UpperDict):
                append(f"{key}: {{")
                for line in str(value).splitlines()[1:-1]:
                    append(f"{line}")
                append("}")
            else:
                if isinstance(value, str):
                    append(f"{key}: '{value}',")
                else:
                    append(f"{key}: {value},")

        result.append("}")

        return "\n".join(result)

    def __setitem__(self, key, value):
        key = key.upper()

        if isinstance(value, dict):
            if key in self.__dict.keys():
                for k, v in value.items():
                    self.__dict[key][k] = v
            else:
                self.__dict[key] = UpperDict(value)
        else:
            if key == "DEBUG":
                value = bool(value)
            elif key == "PORT":
                value = int(value)
            self.__dict[key] = value

    def __getitem__(self, key):
        return self.__dict[key.upper()]

    def __setattr__(self, name, value):
        if name == f"_UpperDict__dict":
            return super().__setattr__(name, value)
        raise ConfigError("Modifying the attribute value of Config is not allowed.")

    def __getattr__(self, name):
        try:
            value = self.get(name)
            # if value is None:
            #     raise KeyError()
            return value
        except KeyError:
            raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}'")

    def get(self, key, default=None):
        try:
            return self.__dict[key.upper()]
        except KeyError:
            return default


def _import_environ():
    result = {}
    for key in filter(
        lambda x: x.startswith("INDEX_") and x.upper() in ("DEBUG", "ENV"),
        os.environ.keys()
    ):
        if key.upper() == "DEBUG":
            result[key[6:]] = os.environ.get(key) in ("on", "True")
        else:
            result[key[6:]] = os.environ.get(key)
    return result


class Config(UpperDict):

    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, '_instance'):
            cls._instance = super().__new__(cls, *args, **kwargs)
        return cls._instance

    def __init__(self):
        super().__init__({})
        self.setdefault()
        # read config from `config.json`
        self.import_from_file(os.path.join(self.path, "config.json"))
        # read config from environ
        self.update(_import_environ())

    @property
    def path(self):
        """return os.getcwd()"""
        return os.getcwd()

    def import_from_file(self, jsonfile: str):
        try:
            with open(jsonfile, "r") as file:
                data = json.load(file)
            if not isinstance(data, dict):
                raise ConfigError(f"config must be a dictionary.")

            self.update(data)
        except FileNotFoundError:
            pass

    def setdefault(self):
        """set default value"""

        self["env"] = "dev"
        self["debug"] = False
        self["host"] = "127.0.0.1"
        self["port"] = 4190
        self['log_level'] = "info"
        # url
        self['allow_underline'] = False
        # middleware
        self['allowed_hosts'] = ["*"]
        self['cors_settings'] = {
            "allow_origins": (),
            "allow_methods": ("GET",),
            "allow_headers": (),
            "allow_credentials": False,
            "allow_origin_regex": None,
            "expose_headers": (),
            "max_age": 600,
        }

    def update(self, data: dict):
        for key in data.keys():
            self[key] = data[key]

    def get(self, key, default=None):
        env = super().get(self['env'], {})
        value = env.get(key, None)
        if value is None:
            value = super().get(key, default)
        return value


logger.setLevel(LOG_LEVELS[Config().log_level])



from .concurrency import complicating


class MiddlewareMixin:
    def __init__(self, get_response):
        self.get_response = complicating(get_response)

    async def __call__(self, request):
        response = None
        if hasattr(self, 'process_request'):
            response = await complicating(self.process_request)(request)

        if response is None:
            response = await self.get_response(request)

        if hasattr(self, 'process_response'):
            response = await complicating(self.process_response)(request, response)
        return response



import asyncio
import functools

from starlette.concurrency import run_in_threadpool


def complicating(func):
    """
    always return a coroutine function
    """
    if asyncio.iscoroutinefunction(func):
        return func

    @functools.wraps(func)
    async def wrapper(*args, **kwargs):
        result = await run_in_threadpool(func, *args, **kwargs)
        if asyncio.iscoroutine(result):
            return await result
        return result
    return wrapper



import os
import re
import sys
import logging
import subprocess
from multiprocessing import cpu_count

import click
import uvicorn

from . import app, Config

logger = logging.getLogger(__name__)
config = Config()


def execute(command: str):
    click.echo("Execute command: ", nl=False)
    click.secho(command, fg="green")
    os.system(command)


@click.group()
def main():
    pass


@main.command(help='use only uvicorn')
def dev():
    uvicorn.run(
        app,
        host=config.HOST,
        port=config.PORT,
        log_level=config.LOG_LEVEL,
        debug=config.DEBUG,
        lifespan="on"
    )


@main.command(help="deploy by gunicorn")
@click.option("--workers", "-w", default=cpu_count())
@click.option("--daemon", "-d", default=False, is_flag=True)
@click.argument("method")
def gunicorn(workers, daemon, method):
    if sys.platform != "linux":
        raise RuntimeError("gunicorn can only run on Linux.")
    if method == "start":
        command = (
            f"gunicorn -k uvicorn.workers.UvicornWorker"
            f" --bind {config.HOST}:{config.PORT}"
            f" --chdir {config.path}"
            f" --pid {os.path.join(config.path, '.pid')}"
            f" --log-level {config.LOG_LEVEL}"
            f"{' -D --log-file log.index' if daemon else ''}"
            f" -w {workers}"
            f" index:app"
        )
        execute(command)
    elif method == "stop":
        execute("kill -TERM `cat .pid`")
    elif method == "reload":
        execute("kill -HUP `cat .pid`")



import re
import os
import logging
import threading
import importlib

from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer

from .config import Config

logger = logging.getLogger(__name__)
config = Config()


class ImportTypeError(Exception):

    def __init__(self, position: str, sentence: str):
        self.position = position
        self.sentence = sentence


class CheckImport:

    def __init__(self):
        self.pattern = re.compile("from (?P<path>.*?) import ")

    def check(self, filepath: str) -> (int, str):
        """
        check `from ... import ...` in file

        https://docs.python.org/zh-cn/3/library/importlib.html#importlib.reload
        """
        with open(filepath, encoding="UTF-8") as file:
            for index, line in enumerate(file.readlines()):
                for path in self.pattern.findall(line):
                    abspath = os.path.join(config.path, path.replace(".", "/")) + ".py"
                    if os.path.isfile(abspath):
                        return index, line.strip()
        return 0, ""

    def __call__(self, path: str) -> bool:
        flag = True
        for root, dirnames, filenames in os.walk(path):
            for filename in filenames:
                if not filename.endswith(".py"):
                    continue
                relpath = os.path.relpath(os.path.join(root, filename), path).replace("\\", "/")
                # check import
                error_line_num, error_sentence = self.check(os.path.join(root, filename))
                if error_line_num > 0:
                    flag = False
                    e = ImportTypeError(f"{relpath} line {error_line_num}", error_sentence)
                    logger.warning(f"Check import type error in {e.position}: '{e.sentence}'")
                    # Preloading
                importlib.import_module(relpath.replace("/", ".")[:-3])
        return flag


class MonitorFileEventHandler(FileSystemEventHandler):

    def dispatch(self, event):
        if not event.src_path.endswith(".py"):
            return
        event.filepath = os.path.relpath(event.src_path, Config().path).replace("\\", "/")[:-3]
        if event.filepath.endswith("/__init__"):
            event.filepath = event.filepath[:-len("/__init__")]
        return super().dispatch(event)

    def on_modified(self, event):
        module_path = ".".join(event.filepath.split("/"))
        logger.debug(f"reloading {event.filepath} as {module_path}")

        def reload():
            module = importlib.import_module(module_path)
            importlib.reload(module)
        threading.Thread(target=reload, daemon=True).start()

    def on_created(self, event):
        module_path = ".".join(event.filepath.split("/"))
        logger.debug(f"loading {event.filepath} as {module_path}")
        threading.Thread(target=importlib.import_module, args=(module_path, ), daemon=True).start()


class MonitorFile:
    def __init__(self, path: str):
        self.observer = Observer()
        self.observer.schedule(MonitorFileEventHandler(), path, recursive=True)
        self.observer.start()

    def __del__(self):
        """drop observer"""
        # self.observer.stop()
        # self.observer.join()



import os
import sys
import json
import logging
import importlib

from starlette.applications import Starlette
from starlette.staticfiles import StaticFiles
from starlette.responses import RedirectResponse
from starlette.middleware.trustedhost import TrustedHostMiddleware
from starlette.middleware.cors import CORSMiddleware
from starlette.exceptions import HTTPException

from .config import Config
from .responses import automatic
from .autoreload import MonitorFile, CheckImport

logger = logging.getLogger(__name__)
config = Config()

sys.path.insert(0, config.path)

app = Starlette(debug=config.DEBUG)

# middleware
app.add_middleware(
    TrustedHostMiddleware,
    allowed_hosts=config.ALLOWED_HOSTS
)
app.add_middleware(
    CORSMiddleware,
    allow_origins=config.CORS_SETTINGS.allow_origins,
    allow_methods=config.CORS_SETTINGS.allow_methods,
    allow_headers=config.CORS_SETTINGS.allow_headers,
    allow_credentials=config.CORS_SETTINGS.allow_credentials,
    allow_origin_regex=config.CORS_SETTINGS.allow_origin_regex,
    expose_headers=config.CORS_SETTINGS.expose_headers,
    max_age=config.CORS_SETTINGS.max_age,
)


@app.on_event('startup')
def startup():
    # monitor file event
    if CheckImport()(config.path):
        monitorfile = MonitorFile(config.path)

    # static & template
    os.makedirs(os.path.join(config.path, "statics"), exist_ok=True)
    os.makedirs(os.path.join(config.path, "templates"), exist_ok=True)


@app.on_event('shutdown')
async def shutdown():
    pass


app.mount('/static', StaticFiles(directory="statics"))


@app.route("/")
def index(request):
    return RedirectResponse("/index.py", status_code=301)


@app.route('/favicon.ico')
def favicon(request):
    return RedirectResponse("/static/favicon.ico")


@app.route("/{filepath:path}.py", methods=['get', 'post', 'put', 'patch', 'delete', 'head', 'options', 'trace'])
async def http(request):
    filepath = request.path_params['filepath']
    # Google SEO
    if not config.ALLOW_UNDERLINE:
        if "_" in filepath:
            return RedirectResponse(f'/{filepath.replace("_", "-")}.py', status_code=301)
        filepath = filepath.strip(".").replace("-", "_")

    # judge python file
    abspath = os.path.join(config.path, "views", filepath + ".py")
    if not os.path.exists(abspath):
        raise HTTPException(404)

    pathlist = ['views'] + filepath.split("/")

    # find http handler
    module_path = ".".join(pathlist)
    module = importlib.import_module(module_path)
    try:
        get_response = module.HTTP()
    except AttributeError:
        raise HTTPException(404)

    # call middleware
    for deep in range(len(pathlist), 0, -1):
        try:
            module = importlib.import_module(".".join(pathlist[:deep]))
            get_response = module.Middleware(get_response)
            logger.debug(f"Call middleware in {module}")
        except AttributeError:
            continue

    # get response
    resp = await get_response(request)
    if not isinstance(resp, tuple):
        resp = (resp,)
    return automatic(*resp)



import logging

from starlette.responses import Response
from starlette.requests import Request

from .concurrency import complicating

logger = logging.getLogger(__name__)


class NoMixedCaseMeta(type):
    def __new__(cls, clsname, bases, clsdict):
        for name in clsdict:
            if name.lower() != name:
                raise TypeError('Bad attribute name: ' + name)
        return super().__new__(cls, clsname, bases, clsdict)


class View(metaclass=NoMixedCaseMeta):
    http_method_names = ['get', 'post', 'put', 'patch', 'delete', 'head', 'options', 'trace']

    def __init__(self):
        pass

    async def __call__(self, request: Request):
        # Try to dispatch to the right method; if a method doesn't exist,
        # defer to the error handler. Also defer to the error handler if the
        # request method isn't on the approved list.
        self.request = request

        if self.request.method.lower() in self.http_method_names:
            handler = getattr(self, self.request.method.lower(), self.http_method_not_allowed)
        else:
            handler = self.http_method_not_allowed
        return await complicating(handler)()

    def http_method_not_allowed(self):
        logger.warning(f'Method Not Allowed ({self.request.method}): {self.request.url.path}')
        return Response(status_code=405, headers={
            "Allow": ', '.join(self.allowed_methods())
        })

    def options(self):
        """Handle responding to requests for the OPTIONS HTTP verb."""
        return Response(headers={
            "Allow": ', '.join(self.allowed_methods()),
            "Content-Length": "0"
        })

    def allowed_methods(self):
        return [m.upper() for m in self.http_method_names if hasattr(self, m)]
